from abc import ABC, abstractmethod
import logging
import os
import queue
import threading
import time
from typing import Callable
import requests
from okahu_apptrace.constants import AWS_LAMBDA_ENV_NAME

logger = logging.getLogger(__name__)
LAMBDA_EXTENSION_NAME = "AsyncProcessorOkahu"


class ExportTaskProcessor(ABC):
    
    @abstractmethod
    def start(self):
        return
    
    @abstractmethod
    def stop(self):
        return
    
    @abstractmethod
    def queue_task(self, async_task:Callable[[Callable, any],any] =None, args: any=None):
        return

class LambdaExportTaskProcessor(ExportTaskProcessor):
    
    def __init__(
        self,
        span_check_interval_seconds: int = 2,
        max_time_allowed_seconds: int = 30):
        # An internal queue used by the handler to notify the extension that it can
        # start processing the async task.
        self.async_tasks_queue = queue.Queue()
        self.span_check_interval = span_check_interval_seconds
        self.max_time_allowed = max_time_allowed_seconds
    
    def is_aws_lambda_environment():
        return AWS_LAMBDA_ENV_NAME in os.environ
    
    def start(self):
        self._start_async_processor()
    
    def stop(self):
        return
    
    def queue_task(self, async_task=None, args=None):
        self.async_tasks_queue.put((async_task, args))
    
    def _start_async_processor(self):
        # Register internal extension
        logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Registering with Lambda service...")
        response = requests.post(
            url=f"http://{os.environ['AWS_LAMBDA_RUNTIME_API']}/2020-01-01/extension/register",
            json={'events': ['INVOKE']},
            headers={'Lambda-Extension-Name': LAMBDA_EXTENSION_NAME}
        )
        ext_id = response.headers['Lambda-Extension-Identifier']
        logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Registered with ID: {ext_id}")

        
        def process_tasks():
            while True:
                # Call /next to get notified when there is a new invocation and let
                # Lambda know that we are done processing the previous task.

                logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Waiting for invocation...")
                response = requests.get(
                    url=f"http://{os.environ['AWS_LAMBDA_RUNTIME_API']}/2020-01-01/extension/event/next",
                    headers={'Lambda-Extension-Identifier': ext_id},
                    timeout=None
                )
                root_span_found = False
                # all values in seconds
                total_time_elapsed = 0
                while root_span_found is False and total_time_elapsed < self.max_time_allowed:
                    logger.debug(response.json())
                    # Get next task from internal queue
                    logger.info(f"[{LAMBDA_EXTENSION_NAME}] Async thread running, waiting for task from handler")
                    while self.async_tasks_queue.empty() is False :
                        async_task, args = self.async_tasks_queue.get()
                        # check if root span task is present in queue
                        if 'batch' in args:
                            for span in args["batch"]:
                                if(span["parent_id"] == "None"):
                                    root_span_found = True
                        
                        if async_task is None:
                            # No task to run this invocation
                            logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Received null task. Ignoring.")
                        else:
                            # Invoke task
                            logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Received async task from handler. Starting task.")
                            async_task(args)
                    total_time_elapsed+=self.span_check_interval
                    time.sleep(self.span_check_interval)
                
                logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Finished processing task. total_time_elapsed: {total_time_elapsed}, root_span_found: {root_span_found}.")

        # Start processing extension events in a separate thread
        threading.Thread(target=process_tasks, daemon=True, name=LAMBDA_EXTENSION_NAME).start() 


