# Copyright (C) Okahu Inc 2023-2024. All rights reserved

import logging
import os
from opentelemetry.trace import Span

from okahu_apptrace.constants import azure_service_map, aws_service_map, infra_identifier_env_list
logger = logging.getLogger(__name__)

class Config:
    exception_logger = None

def set_span_attribute(span, name, value):
    if value is not None:
        if value != "":
            span.set_attribute(name, value)
    return


def dont_throw(func):
    """
    A decorator that wraps the passed in function and logs exceptions instead of throwing them.

    @param func: The function to wrap
    @return: The wrapper function
    """
    # Obtain a logger specific to the function's module
    logger = logging.getLogger(func.__module__)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.warning("Failed to execute %s, error: %s", func.__name__, str(e))
            if Config.exception_logger:
                Config.exception_logger(e)

    return wrapper

def with_tracer_wrapper(func):
    """Helper for providing tracer for wrapper functions."""

    def _with_tracer(tracer, to_wrap):
        def wrapper(wrapped, instance, args, kwargs):
            return func(tracer, to_wrap, wrapped, instance, args, kwargs)

        return wrapper

    return _with_tracer

def resolve_from_alias(map, alias):
    """Find a alias that is not none from list of aliases"""

    for i in alias:
        if i in map.keys():
            return map[i]
    return None

def update_span_with_infra_name(span: Span, span_key: str):
    for service_map in [azure_service_map, aws_service_map]:
        set_infra_attribute(span, span_key, service_map)

def set_infra_attribute(span: Span, span_key: str, service_map: dict[str, str]):
    for key,val  in service_map.items():
        if key in os.environ:
            span.set_attribute(span_key, val)

def update_span_with_infra_identifier_name(span: Span, span_key: str):
    for val  in infra_identifier_env_list:
        if val in os.environ:
            span.set_attribute(span_key, os.environ[val])
